<?php
/**
 * Perforce Swarm
 *
 * @copyright   2013-2019 Perforce Software. All rights reserved.
 * @license     Please see LICENSE.txt in top-level readme folder of this distribution.
 * @version     2018.3/1751405
 */

namespace Api\Controller;

use Api\AbstractApiController;
use Application\Permissions\Exception\ForbiddenException;
use Projects\Model\Project;
use Workflow\Manager;
use Zend\Http\Response;
use Zend\View\Model\JsonModel;
use Workflow\Model\IWorkflow;
use Workflow\Model\Workflow;
use Workflow\Filter\Workflow as Filter;
use Record\Exception\NotFoundException;

/**
 * Controller for querying, creating and updating workflows.
 *
 * @SWG\Resource(
 *   apiVersion="v9",
 *   basePath="/api/v9/"
 * )
 */
class WorkflowsController extends AbstractApiController
{
    /**
     * @SWG\Api(
     *     path="workflows/",
     *     @SWG\Operation(
     *         method="POST",
     *         summary="Create a workflow",
     *         notes="Create a new workflow.",
     *         nickname="createWorkflow",
     *         @SWG\Parameter(
     *             name="name",
     *             description="The workflow name. Will be compared against other workflows and rejected if not unique",
     *             paramType="form",
     *             type="string",
     *             required=true
     *         ),
     *         @SWG\Parameter(
     *             name="description",
     *             description="Description for the new workflow",
     *             paramType="form",
     *             type="string",
     *             required=false
     *         ),
     *         @SWG\Parameter(
     *             name="shared",
     *             description="Whether this workflow is shared for other users that do not own it.
     *                          Defaults to not shared",
     *             paramType="form",
     *             type="boolean",
     *             required=false
     *         ),
     *         @SWG\Parameter(
     *             name="owners",
     *             description="A list owners for the workflow. Can be users or group names
     *                          (prefixed with swarm-group-). Users and group names must exist or the
     *                          workflow will be rejected",
     *             paramType="form",
     *             type="array",
     *             required=false,
     *             @SWG\Items("string")
     *         ),
     *         @SWG\Parameter(
     *             name="on_submit",
     *             description="Data for rules when changes are submitted. Valid values for with_review are no_checking,
     *                          approved, strict. Valid values for without review are no_checking, auto_create, reject",
     *             paramType="form",
     *             type="array",
     *             required=false
     *         ),
     *         @SWG\Parameter(
     *             name="end_rules",
     *             description="Data for rules when changes are submitted. Valid values are no_checking, no_revision.",
     *             paramType="form",
     *             type="array",
     *             required=false
     *         ),
     *         @SWG\Parameter(
     *             name="auto_approve",
     *             description="Data for rules when changes are submitted. Valid values are votes, never.",
     *             paramType="form",
     *             type="array",
     *             required=false
     *         ),
     *         @SWG\Parameter(
     *             name="counted_votes",
     *             description="Data for rules when counting votes up. Valid values are anyone, members.",
     *             paramType="form",
     *             type="string",
     *             required=false
     *         ),
     *         @SWG\ResponseMessage(code=400, message="Bad Request"),
     *         @SWG\ResponseMessage(code=403, message="Unauthorized")
     *     )
     * )
     * @apiUsageExample Create a workflow
     *
     *   ```bash
     *   curl -u "<user>:<password>" \
     *        -H "Content-Type: application/json" \
     *        -X POST \
     *        -d '"name": "myWorkflow", \
     *            "owners": ["user1", "user2"], \
     *            "description": "A description", \
     *            "shared": true, \
     *            "on_submit":{ \
     *              "with_review":{"rule": "no_checking"},' \
     *              "without_review":{"rule": "no_checking"}}' \
     *            "auto_approve": "never", \
     *            "counted_votes": "members", \
     *        "https://my-swarm-host/api/v9/workflows"
     *   ```
     *
     *   JSON Response:
     *
     *   ```json
     * {
     *      "workflow": {
     *          "id": "1",
     *          "name": "myWorkflow",
     *          "description": "A description",
     *          "on_submit": {
     *              "with_review": {
     *                  "rule": "no_checking"
     *              },
     *              "without_review": {
     *                  "rule": "no_checking"
     *              }
     *          },
     *          "auto_approve": {
     *              "rule": "never"
     *          },
     *          "counted_votes": {
     *              "rule": "members"
     *          },
     *          "shared": "true",
     *          "owners": [
     *              "user1",
     *              "user2"
     *          ]
     *      }
     * }
     *   ```
     *
     * @apiSuccessExample Successful Response:
     *     HTTP/1.1 200 OK
     *
     * {
     *      "workflow": {
     *          "id": "1",
     *          "name": "myWorkflow",
     *          "description": "A description",
     *          "on_submit": {
     *              "with_review": {
     *                  "rule": "no_checking"
     *              },
     *              "without_review": {
     *                  "rule": "no_checking"
     *              }
     *          },
     *          "auto_approve": {
     *              "rule": "never"
     *          },
     *          "counted_votes": {
     *              "rule": "members"
     *          },
     *          "shared": "true",
     *          "owners": [
     *              "user1",
     *              "user2"
     *          ]
     *      }
     * }
     *
     * @param mixed $data
     * @return JsonModel
     */
    public function create($data)
    {
        $services = $this->getServiceLocator();
        try {
            $services->get('config_check')->enforce(IWorkflow::WORKFLOW, 'authenticated');
        } catch (ForbiddenException $e) {
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_501);
            return $this->buildErrorResponse($e);
        }
        $p4Admin = $services->get('p4_admin');
        $data    = $this->normaliseData($data);
        $filter  = new Filter($p4Admin);
        $filter->setMode(Filter::MODE_ADD)->setData($data);
        return $this->buildChangeResponse(new Workflow($p4Admin), $filter);
    }

    /**
     * @SWG\Api(
     *     path="workflows/{id}",
     *     @SWG\Operation(
     *         method="PATCH",
     *         summary="Patch a workflow",
     *         notes="Patch a workflow.",
     *         nickname="patchWorkflow",
     *         @SWG\Parameter(
     *             name="id",
     *             description="The id of the workflow being patched",
     *             paramType="form",
     *             type="string",
     *             required=true
     *         ),
     *         @SWG\Parameter(
     *             name="name",
     *             description="The workflow name. Will be compared against other workflows and rejected if not unique",
     *             paramType="form",
     *             type="string",
     *             required=false
     *         ),
     *         @SWG\Parameter(
     *             name="description",
     *             description="Description for the new workflow",
     *             paramType="form",
     *             type="string",
     *             required=false
     *         ),
     *         @SWG\Parameter(
     *             name="shared",
     *             description="Whether this workflow is shared for other users that do not own it.
     *                          Defaults to not shared",
     *             paramType="form",
     *             type="boolean",
     *             required=false
     *         ),
     *         @SWG\Parameter(
     *             name="owners",
     *             description="A list owners for the workflow. Can be users or group names
     *                          (prefixed with swarm-group-). Users and group names must exist or the
     *                          workflow will be rejected",
     *             paramType="form",
     *             type="array",
     *             required=false,
     *             @SWG\Items("string")
     *         ),
     *         @SWG\Parameter(
     *             name="on_submit",
     *             description="Data for rules when changes are submitted. Valid values for with_review are no_checking,
     *                          approved, strict. Valid values for without review are no_checking, auto_create, reject",
     *             paramType="form",
     *             type="array",
     *             required=false
     *         ),
     *         @SWG\Parameter(
     *             name="end_rules",
     *             description="Data for rules when changes are submitted. Valid values are no_checking, no_revision.",
     *             paramType="form",
     *             type="array",
     *             required=false
     *         ),
     *         @SWG\Parameter(
     *             name="auto_approve",
     *             description="Data for rules when changes are submitted. Valid values are votes, never.",
     *             paramType="form",
     *             type="array",
     *             required=false
     *         ),
     *         @SWG\Parameter(
     *             name="counted_votes",
     *             description="Data for rules when counting votes up. Valid values are anyone, members.",
     *             paramType="form",
     *             type="string",
     *             required=false
     *         ),
     *         @SWG\ResponseMessage(code=400, message="Bad Request"),
     *         @SWG\ResponseMessage(code=403, message="Unauthorized")
     *     )
     * )
     * @apiUsageExample Update a workflow
     *
     *   ```bash
     *   curl -u "username:password" \
     *        -X PATCH \
     *        -d "name=myWorkflow" \
     *        -d "description=A description" \
     *        "https://my-swarm-host/api/v9/workflows/1"
     *   ```
     *
     *   JSON Response:
     *
     *   ```json
     * {
     *      "workflow": {
     *          "id": "1",
     *          "name": "myWorkflow",
     *          "description": "A description",
     *          "on_submit": {
     *              "with_review": {
     *                  "rule": "no_checking"
     *              },
     *              "without_review": {
     *                  "rule": "no_checking"
     *              }
     *          },
     *          "auto_approve": {
     *              "rule": "never"
     *          },
     *          "counted_votes": {
     *              "rule": "anyone"
     *          },
     *          "shared": "true",
     *          "owners": [
     *              "user1",
     *              "user2"
     *          ]
     *      }
     * }
     *   ```
     *
     * @apiSuccessExample Successful Response:
     *     HTTP/1.1 200 OK
     *
     * {
     *      "workflow": {
     *          "id": "1",
     *          "name": "myWorkflow",
     *          "description": "A description",
     *          "on_submit": {
     *              "with_review": {
     *                  "rule": "no_checking"
     *              },
     *              "without_review": {
     *                  "rule": "no_checking"
     *              }
     *          },
     *          "auto_approve": {
     *              "rule": "never"
     *          },
     *          "counted_votes": {
     *              "rule": "anyone"
     *          },
     *          "shared": "true",
     *          "owners": [
     *              "user1",
     *              "user2"
     *          ]
     *      }
     * }
     *
     * @param string        $workflowId     The id of the workflow being patched
     * @param array         $data           Data to patch
     * @return JsonModel
     */
    public function patch($workflowId, $data)
    {
        $error = null;
        $model = null;
        try {
            $services = $this->getServiceLocator();
            $services->get('config_check')->enforce(IWorkflow::WORKFLOW, 'authenticated');
            $p4Admin  = $services->get('p4_admin');
            $p4_user  = $services->get('p4_user');
            $data     = $this->normaliseData($data);
            $workflow = Workflow::fetch($workflowId, $p4Admin);
            if ($workflow->canEdit($p4_user)) {
                $data   = array_replace($workflow->toArray(), $data);
                $filter = new Filter($p4Admin, $workflow);
                $filter->setMode(Filter::MODE_EDIT)->setData($data);
                $model = $this->buildChangeResponse($workflow, $filter);
            } else {
                // No permission - we'll treat this as not found so as to not give an clues
                throw new NotFoundException('Cannot fetch entry. Id does not exist.');
            }
        } catch (NotFoundException $e) {
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_404);
            $error = $e;
        } catch (\InvalidArgumentException $e) {
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_404);
            $error = $e;
        } catch (ForbiddenException $e) {
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_501);
            $error = $e;
        }
        if ($error) {
            $model = $this->buildErrorResponse($error);
        }
        return $model;
    }

    /**
     * Returns the array passed with only relevant workflow values set.
     * @param array $data data to normalise
     * @return array the normalised array
     */
    private function normaliseData(array $data)
    {
        // Only allow expected inputs
        return array_intersect_key(
            $data,
            array_flip(
                array(
                    IWorkflow::SHARED,
                    IWorkflow::ON_SUBMIT,
                    IWorkflow::NAME,
                    IWorkflow::DESCRIPTION,
                    IWorkflow::OWNERS,
                    IWorkflow::END_RULES,
                    IWorkflow::AUTO_APPROVE,
                    IWorkflow::COUNTED_VOTES
                )
            )
        );
    }

    /**
     * Builds a response for update requests PUT, PATCH and POST
     * @param IWorkflow     $workflow   the workflow being changed
     * @param Filter        $filter     the filter to use for values
     * @return JsonModel
     */
    private function buildChangeResponse($workflow, Filter $filter)
    {
        if ($filter->isValid()) {
            $workflow->set($filter->getValues())->save();
            $model = $this->prepareSuccessModel(array('workflow' => $workflow->toArray()));
        } else {
            // Unless the filter specifically set forbidden set the response code to the general 400
            if ($this->getResponse()->getStatusCode() !== Response::STATUS_CODE_403) {
                $this->getResponse()->setStatusCode(Response::STATUS_CODE_400);
            }
            $jsonModel = new JsonModel(
                array(
                    'isValid'  => false,
                    'messages' => $filter->getMessages(),
                    'workflow' => null,
                )
            );
            $model     = $this->prepareErrorModel($jsonModel);
        }
        return $model;
    }

    /**
     * @SWG\Api(
     *     path="workflows/{id}",
     *     @SWG\Operation(
     *         method="DELETE",
     *         summary="Delete a workflow",
     *         notes="Delete a workflow for the provided id. This call must be authenticated and the user must
     *                have permission to edit the workflow. If the workflow is in use it cannot be deleted and
     *                and error message will be returned",
     *         nickname="updateWorkflow",
     *         @SWG\Parameter(
     *             name="id",
     *             description="The id of the workflow being deleted",
     *             paramType="form",
     *             type="string",
     *             required=true
     *         ),
     *         @SWG\ResponseMessage(code=400, message="Bad Request"),
     *         @SWG\ResponseMessage(code=403, message="Unauthorized")
     *     )
     * )
     * @apiUsageExample Delete a workflow not in use
     *
     *   ```bash
     *   curl -u "<user>:<password>" \
     *        -X DELETE \
     *        "https://my-swarm-host/api/v9/workflows/1"
     *   ```
     *
     * @apiSuccessExample Successful Response:
     *     HTTP/1.1 200 OK
     *
     * {
     *      "isValid": true,
     *      "messages": [
     *          "Workflow [1] was deleted"
     *      ]
     * }
     *
     * @apiUsageExample Delete a workflow that is in use on public projects
     *
     *   ```bash
     *   curl -u "<user>:<password>" \
     *        -X DELETE \
     *        "https://my-swarm-host/api/v9/workflows/1"
     *   ```
     *
     * @apiSuccessExample Successful Response:
     *     HTTP/1.1 200 OK
     *
     * {
     *      "isValid": false,
     *      "messages": [
     *          "Cannot delete workflow [1], it is in use on project [project1, project2]"
     *      ]
     * }
     *
     * @apiUsageExample Delete a workflow that is in use on a public and some private projects
     *
     *   ```bash
     *   curl -u "<user>:<password>" \
     *        -X DELETE \
     *        "https://my-swarm-host/api/v9/workflows/1"
     *   ```
     *
     * @apiSuccessExample Successful Response:
     *     HTTP/1.1 200 OK
     *
     * {
     *      "isValid": false,
     *      "messages": [
     *          "Cannot delete workflow [1], it is in use on project [project1] and others"
     *      ]
     * }
     *
     * @apiUsageExample Delete a workflow that is in use on only private projects
     *
     *   ```bash
     *   curl -u "<user>:<password>" \
     *        -X DELETE \
     *        "https://my-swarm-host/api/v9/workflows/1"
     *   ```
     *
     * @apiSuccessExample Successful Response:
     *     HTTP/1.1 200 OK
     *
     * {
     *      "isValid": false,
     *      "messages": [
     *          "Cannot delete workflow [1], it is in use"
     *      ]
     * }
     *
     * @param mixed $workflowId
     * @return mixed|null|JsonModel
     * @throws \Exception
     */
    public function delete($workflowId)
    {
        $error = null;
        $model = null;
        try {
            $services = $this->getServiceLocator();
            $services->get('config_check')->enforce(IWorkflow::WORKFLOW, 'authenticated');
            $p4Admin  = $services->get('p4_admin');
            $p4_user  = $services->get('p4_user');
            $workflow = Workflow::fetch($workflowId, $p4Admin);
            if ($workflow->canEdit($p4_user)) {
                $projects   = Project::fetchAll(array(Project::FIELD_WORKFLOW => (string) $workflowId), $p4Admin);
                $allCount   = count($projects);
                $translator = $services->get('translator');
                if (!$projects || $allCount === 0) {
                    $workflow->delete();
                    $model = new JsonModel(
                        array(
                            'isValid'  => true,
                            'messages' => array($translator->t('Workflow [%s] was deleted', array($workflowId))),
                        )
                    );
                } else {
                    $projects = $services->get('projects_filter')->filter($projects);
                    $inUseOn  = array_map(
                        function ($project) {
                            return $project->getName();
                        },
                        $projects->getArrayCopy()
                    );
                    $inUseCount = count($inUseOn);
                    if ($inUseCount === 0) {
                        $message = $translator->t('Cannot delete workflow [%s], it is in use', array($workflowId));
                    } else {
                        $andOthers = $inUseCount === $allCount ? '' : ' and others';
                        $message   = $translator->tp(
                            'Cannot delete workflow [%s], it is in use on project [%s]' . $andOthers,
                            'Cannot delete workflow [%s], it is in use on projects [%s]' . $andOthers,
                            count($inUseOn),
                            array($workflowId, implode(', ', $inUseOn))
                        );
                    }
                    $model = new JsonModel(
                        array(
                            'isValid'  => false,
                            'messages' => array($message),
                        )
                    );
                }
            } else {
                // No permission - we'll treat this as not found so as to not give an clues
                throw new NotFoundException('Cannot fetch entry. Id does not exist.');
            }
        } catch (NotFoundException $e) {
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_404);
            $error = $e;
        } catch (\InvalidArgumentException $e) {
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_404);
            $error = $e;
        } catch (ForbiddenException $e) {
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_501);
            $error = $e;
        }
        if ($error) {
            $model = $this->buildErrorResponse($error);
        }
        return $model;
    }

    /**
     * Build a response based on an exception
     * @param mixed $error  the exception
     * @return JsonModel
     */
    private function buildErrorResponse($error)
    {
        return new JsonModel(
            array(
                'isValid'  => false,
                'messages' => array($error->getMessage()),
            )
        );
    }

    /**
     * @SWG\Api(
     *     path="workflows/{id}",
     *     @SWG\Operation(
     *         method="PUT",
     *         summary="Update a workflow",
     *         notes="Update a workflow. All values should be provided in the request. If not provided any missing
     *                values are reverted to default.",
     *         nickname="updateWorkflow",
     *         @SWG\Parameter(
     *             name="id",
     *             description="The id of the workflow being patched",
     *             paramType="form",
     *             type="string",
     *             required=true
     *         ),
     *         @SWG\Parameter(
     *             name="name",
     *             description="The workflow name. Will be compared against other workflows and rejected if not unique",
     *             paramType="form",
     *             type="string",
     *             required=true
     *         ),
     *         @SWG\Parameter(
     *             name="owners",
     *             description="A list owners for the workflow. Can be users or group names
     *                          (prefixed with swarm-group-). Users and group names must exist or the
     *                          workflow will be rejected",
     *             paramType="form",
     *             type="array",
     *             required=true,
     *             @SWG\Items("string")
     *         ),
     *         @SWG\Parameter(
     *             name="description",
     *             description="Description for the new workflow",
     *             paramType="form",
     *             type="string",
     *             required=false
     *         ),
     *         @SWG\Parameter(
     *             name="shared",
     *             description="Whether this workflow is shared for other users that do not own it.
     *                          Defaults to not shared",
     *             paramType="form",
     *             type="boolean",
     *             required=false
     *         ),
     *         @SWG\Parameter(
     *             name="on_submit",
     *             description="Data for rules when changes are submitted. Valid values for with_review are no_checking,
     *                          approved, strict. Valid values for without review are no_checking, auto_create, reject",
     *             paramType="form",
     *             type="array",
     *             required=false
     *         ),
     *         @SWG\Parameter(
     *             name="end_rules",
     *             description="Data for rules when changes are submitted. Valid values are no_checking, no_revision.",
     *             paramType="form",
     *             type="array",
     *             required=false
     *         ),
     *         @SWG\Parameter(
     *             name="auto_approve",
     *             description="Data for rules when changes are submitted. Valid values are votes, never.",
     *             paramType="form",
     *             type="array",
     *             required=false
     *         ),
     *         @SWG\Parameter(
     *             name="counted_votes",
     *             description="Data for rules when counting votes up. Valid values are anyone, members.",
     *             paramType="form",
     *             type="string",
     *             required=false
     *         ),
     *         @SWG\ResponseMessage(code=400, message="Bad Request"),
     *         @SWG\ResponseMessage(code=403, message="Unauthorized")
     *     )
     * )
     * @apiUsageExample Update a workflow
     *
     *   ```bash
     *   curl -u "<user>:<password>" \
     *        -H "Content-Type: application/json" \
     *        -X POST \
     *        -d '"name": "myWorkflow", \
     *            "owners": ["user1", "user2"], \
     *            "description": "A description", \
     *            "shared": true, \
     *            "on_submit":{ \
     *              "with_review":{"rule": "no_checking"},' \
     *              "without_review":{"rule": "no_checking"}}' \
     *        "https://my-swarm-host/api/v9/workflows/1"
     *   ```
     *
     *   JSON Response:
     *
     *   ```json
     * {
     *      "workflow": {
     *          "id": "1",
     *          "name": "myWorkflow",
     *          "description": "A description",
     *          "on_submit": {
     *              "with_review": {
     *                  "rule": "no_checking"
     *              },
     *              "without_review": {
     *                  "rule": "no_checking"
     *              }
     *          },
     *          "auto_approve": {
     *              "rule": "votes"
     *          },
     *          "counted_votes": {
     *              "rule": "anyone"
     *          },
     *          "shared": "true",
     *          "owners": [
     *              "user1",
     *              "user2"
     *          ]
     *      }
     * }
     *   ```
     *
     * @apiSuccessExample Successful Response:
     *     HTTP/1.1 200 OK
     *
     * {
     *      "workflow": {
     *          "id": "1",
     *          "name": "myWorkflow",
     *          "description": "A description",
     *          "on_submit": {
     *              "with_review": {
     *                  "rule": "no_checking"
     *              },
     *              "without_review": {
     *                  "rule": "no_checking"
     *              }
     *          },
     *          "auto_approve": {
     *              "rule": "votes"
     *          },
     *          "counted_votes": {
     *              "rule": "anyone"
     *          },
     *          "shared": "true",
     *          "owners": [
     *              "user1",
     *              "user2"
     *          ]
     *      }
     * }
     *
     * @param string        $workflowId     The id of the workflow being updated
     * @param array         $data           Data to patch
     * @return JsonModel
     * @throws \Application\Config\ConfigException
     */
    public function update($workflowId, $data)
    {
        $error = null;
        $model = null;
        try {
            $services = $this->getServiceLocator();
            $services->get('config_check')->enforce(IWorkflow::WORKFLOW, 'authenticated');
            $p4Admin = $services->get('p4_admin');
            $p4_user = $services->get('p4_user');
            $manager = new Manager($services);
            $data    = $this->normaliseData($data);
            // PUT should have provided all values the data just replaces defaults
            $data     = array_replace($manager->getDefaults(), $data);
            $workflow = Workflow::fetch($workflowId, $p4Admin);
            if ($workflow->canEdit($p4_user)) {
                $filter = new Filter($p4Admin, $workflow);
                $filter->setMode(Filter::MODE_EDIT)->setData($data);
                $model = $this->buildChangeResponse($workflow, $filter);
            } else {
                // No permission - we'll treat this as not found so as to not give an clues
                throw new NotFoundException('Cannot fetch entry. Id does not exist.');
            }
        } catch (NotFoundException $e) {
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_404);
            $error = $e;
        } catch (\InvalidArgumentException $e) {
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_404);
            $error = $e;
        } catch (ForbiddenException $e) {
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_501);
            $error = $e;
        }
        if ($error) {
            $model = $this->buildErrorResponse($error);
        }
        return $model;
    }

    /**
     * Gets workflows
     * @SWG\Api(
     *     path="workflows/",
     *     @SWG\Operation(
     *         method="GET",
     *         summary="Gets workflows",
     *         notes="Get all workflows",
     *         nickname="getAllWorkflows",
     *         @SWG\Parameter(
     *             name="fields",
     *             description="An optional comma-separated list (or array) of fields to show for each workflow.
     *                          Omitting this parameter or passing an empty value shows all fields.",
     *             paramType="query",
     *             type="string",
     *             required=false
     *         ),
     *         @SWG\Parameter(
     *             name="noCache",
     *             description="If provided and has a value of 'true' a query will always be performed and
     *                          the cache of workflows is ignored. Otherwise the cache will be used if it
     *                          exists.",
     *             paramType="query",
     *             type="boolean",
     *             required=false
     *         )
     *     )
     * )
     *
     * @apiUsageExample Get a list of workflows
     *
     *   ```bash
     *   curl -u "username:password" \
     *        "https://my-swarm-host/api/v9/workflows"
     *   ```
     *
     *   JSON Response:
     *
     *   ```json
     * {
     *      "workflows": [{
     *          "id": "1",
     *          "name": "myWorkflow",
     *          "description": "A description",
     *          "on_submit": {
     *              "with_review": {
     *                  "rule": "no_checking"
     *              },
     *              "without_review": {
     *                  "rule": "no_checking"
     *              }
     *          },
     *          "auto_approve": {
     *              "rule": "never"
     *          },
     *          "counted_votes": {
     *              "rule": "anyone"
     *          },
     *          "shared": "true",
     *          "owners": [
     *              "user1",
     *              "user2"
     *          ]
     *      },
     *      {
     *          "id": "2",
     *          "name": "myWorkflow 2",
     *          "description": "A description" ,
     *          "on_submit": {
     *              "with_review": {
     *                  "rule": "no_checking"
     *              },
     *              "without_review": {
     *                  "rule": "no_checking"
     *              }
     *          },
     *          "auto_approve": {
     *              "rule": "votes"
     *          },
     *          "counted_votes": {
     *              "rule": "members"
     *          },
     *          "shared": "true",
     *          "owners": [
     *              "user3",
     *              "user4"
     *          ]
     *      }
     *      ]
     * }
     *   ```
     *
     * @apiSuccessExample Successful Response:
     *     HTTP/1.1 200 OK
     *
     * {
     *      "workflows": [{
     *          "id": "1",
     *          "name": "myWorkflow",
     *          "description": "A description",
     *          "on_submit": {
     *              "with_review": {
     *                  "rule": "no_checking"
     *              },
     *              "without_review": {
     *                  "rule": "no_checking"
     *              }
     *          },
     *          "auto_approve": {
     *              "rule": "never"
     *          },
     *          "counted_votes": {
     *              "rule": "anyone"
     *          },
     *          "shared": "true",
     *          "owners": [
     *              "user1",
     *              "user2"
     *          ]
     *      },
     *      {
     *          "id": "2",
     *          "name": "myWorkflow 2",
     *          "description": "A description" ,
     *          "on_submit": {
     *              "with_review": {
     *                  "rule": "no_checking"
     *              },
     *              "without_review": {
     *                  "rule": "no_checking"
     *              }
     *          },
     *          "auto_approve": {
     *              "rule": "votes"
     *          },
     *          "counted_votes": {
     *              "rule": "members"
     *          },
     *          "shared": "true",
     *          "owners": [
     *              "user3",
     *              "user4"
     *          ]
     *      }
     *      ]
     * }
     *
     * @return JsonModel
     */
    public function getList()
    {
        $services = $this->getServiceLocator();
        try {
            $services->get('config_check')->enforce(IWorkflow::WORKFLOW);
        } catch (ForbiddenException $e) {
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_501);
            return $this->buildErrorResponse($e);
        }
        $fields  = $this->getRequest()->getQuery('fields');
        $noCache = $this->getRequest()->getQuery(Workflow::FETCH_NO_CACHE);
        $p4Admin = $services->get('p4_admin');
        $options = array();
        if (isset($noCache) && $noCache === 'true') {
            $options[Workflow::FETCH_NO_CACHE] = true;
        }
        $workflows        = Workflow::fetchAll($options, $p4Admin);
        $limitedWorkflows = array();

        foreach ($workflows->toArray() as $workflow) {
            $limitedWorkflows[] = $this->limitEntityFields($workflow, $fields);
        }
        return $this->prepareSuccessModel(new JsonModel(array('workflows' => $limitedWorkflows)));
    }

    /**
     * Gets a specific workflow by its id
     * @SWG\Api(
     *     path="workflows/{id}",
     *     @SWG\Operation(
     *         method="GET",
     *         summary="Gets a workflow by id",
     *         notes="Gets a workflow by id",
     *         nickname="getWorkflow",
     *         @SWG\Parameter(
     *             name="fields",
     *             description="An optional comma-separated list (or array) of fields to show for each workflow.
     *                          Omitting this parameter or passing an empty value shows all fields.",
     *             paramType="query",
     *             type="string",
     *             required=false
     *         )
     *     )
     * )
     *
     * @apiUsageExample Get a list of workflows
     *
     *   ```bash
     *   curl -u "username:password" \
     *        "https://my-swarm-host/api/v9/workflows/1"
     *   ```
     *
     *   JSON Response:
     *
     *   ```json
     * {
     *      "workflow": {
     *          "id": "1",
     *          "name": "myWorkflow",
     *          "description": "A description",
     *          "on_submit": {
     *              "with_review": {
     *                  "rule": "no_checking"
     *              },
     *              "without_review": {
     *                  "rule": "no_checking"
     *              }
     *          },
     *          "auto_approve": {
     *              "rule": "never"
     *          },
     *          "counted_votes": {
     *              "rule": "anyone"
     *          },
     *          "shared": "true",
     *          "owners": [
     *              "user1",
     *              "user2"
     *          ]
     *      }
     * }
     *   ```
     *
     * @apiSuccessExample Successful Response:
     *     HTTP/1.1 200 OK
     *
     * {
     *      "workflow": {
     *          "id": "1",
     *          "name": "myWorkflow",
     *          "description": "A description",
     *          "on_submit": {
     *              "with_review": {
     *                  "rule": "no_checking"
     *              },
     *              "without_review": {
     *                  "rule": "no_checking"
     *              }
     *          },
     *          "auto_approve": {
     *              "rule": "never"
     *          },
     *          "counted_votes": {
     *              "rule": "anyone"
     *          },
     *          "shared": "true",
     *          "owners": [
     *              "user1",
     *              "user2"
     *          ]
     *      }
     * }
     *
     * @param string $workflowId the id of the workflow
     * @return JsonModel
     */
    public function get($workflowId)
    {
        $workflow = null;
        $error    = null;
        try {
            $services = $this->getServiceLocator();
            $services->get('config_check')->enforce(IWorkflow::WORKFLOW);
            $fields   = $this->getRequest()->getQuery('fields');
            $p4Admin  = $services->get('p4_admin');
            $workflow = Workflow::fetch($workflowId, $p4Admin);
        } catch (NotFoundException $e) {
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_404);
            $error = $e;
        } catch (\InvalidArgumentException $e) {
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_404);
            $error = $e;
        } catch (ForbiddenException $e) {
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_501);
            $error = $e;
        }
        if ($workflow) {
            $this->getResponse()->setStatusCode(Response::STATUS_CODE_200);
            return $this->prepareSuccessModel(
                new JsonModel(
                    array('workflow' => $this->limitEntityFields($workflow->toArray(), $fields))
                )
            );
        } else {
            return $this->buildErrorResponse($error);
        }
    }
}
